/*---------------------------------------------------------------------
 *				evo_eeprom.c
 *---------------------------------------------------------------------
 *
 * David Parkinson
 *
 *---------------------------------------------------------------------
 * Most of this code comes directly from natsemi-diag.c whose header
 * follows.  Everything except the initialisation and
 * EEPROM read/routines have been discarded.
 *   Source: natsemi-diag.c:v2.07 12/17/2002
 *---------------------------------------------------------------------
 * natsemi-diag.c: National Semiconductor DP83815 diagnostic and setup.
 *
 * This is a diagnostic and EEPROM setup program for Ethernet adapters
 * based on the National Semiconductor DP83810 / 83815
 *
 * Written 1998-2002 by Donald Becker.
 * Copyright Donald Becker and Scyld Computing Corporation
 *
 * This software may be used and distributed according to the terms of
 * the GNU General Public License (GPL), incorporated herein by reference.
 * Contact the author for use under other terms.
 *
 * This program must be compiled with "-O"!
 * See the bottom of this file for the suggested compile-command.
 *
 * The author may be reached as becker@scyld.com, or C/O
 * Scyld Computing Corporation
 *  410 Severn Ave., Suite 210
 *  Annapolis MD 21403
 *
 * Support and updates available at
 * http://www.scyld.com/diag/index.html
 * http://scyld.com/expert/mii-status.html
 * http://scyld.com/expert/NWay.html
 *
 * Common-sense licensing statement: Using any portion of this program
 * in your own program means that you must give credit to the original
 * author and release the resulting code under the GPL.
 *---------------------------------------------------------------------*/

static char *version_msg = "evo_eeprom V0.1  04/04/2010\n";
static char *usage_msg   = "Usage: evo_eeprom [-r -[w|W] addr:data]\n"
"  Prints the contents of the serial EEPROM in the Evo T20\n"
"  For the write (-w) <addr> and <date> are hexadecimal numbers\n";
static const char long_usage_msg[] =
"Usage: evo_eeprom [-r -[w|W] addr:data]\n"
"  Prints the contents of the 64-word serial EEPROM in the Evo T20\n"
"  and optionally updates a location with new data.\n"
"     -[w|W] addr:data  (addr and data are taken as hexadecimal numbers eg '-w 1e:07ff')\n"
"  changes location [addr] to [data] and automatically updates the checksum\n"
"     for -w the program tells you what it's going to do but doesn't actually do the write\n"
"     for -W it does the write\n"
"     -r displays the data in bit-reversed order.\n";

#if !defined(__OPTIMIZE__)
#warning  You must compile this program with the correct options!
#warning  See the last lines of the source file.
#error You must compile this driver with "-O".
#endif

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <getopt.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <ctype.h>
#include <sys/io.h>

extern int natsemi_diag(int vend_id, int dev_id, long ioaddr, int part_idx);

/* Chip-specific flags. Yes, it's grungy to have the enum here. */

/* The table of known chips.
   Because of the bogus /proc/pci interface we must have both the exact
   name from the kernel, a common name and the PCI vendor/device IDs.
   This table is searched in order: place specific entries followed by
   'catch-all' general entries. */
struct pcidev_entry {
	const char *part_name;
	const char *proc_pci_name;
	int vendor, device, device_mask;
	int flags;
	int io_size;
	int (*diag_func)(int vendor_id, int device_id, long ioaddr, int part_idx);
} pcidev_tbl[] = {
	{"NatSemi DP83815", "National Semiconductor DP83815",
	 0x100B, 0x0020, 0xffff, 0, 256, natsemi_diag},
	{ 0, 0, 0, 0},
};

int verbose = 1;
int debug   = 0;
int reverse = 0;
int do_write= 0;
unsigned int ee_addr=0xFF, ee_data=0;

static int scan_proc_pci(int card_num);

int usage( void )
{
    fprintf( stderr, usage_msg );
    return(3);
}

int main(int argc, char **argv)
{
    int card_num = 0;
    int  n;
    char *p;
    // Check the arguments

    n = 0;
    while( ++n < argc )
    {
	p = argv[n];
	if( *p++ != '-' )
	    usage();
	switch(*p++) {
	    case 'D': debug++;   break;
	    case 'r': reverse++; break;
	    case 'v': verbose++; break;
	    case 'w':
	    case 'W':
		if( *(p-1) == 'W' )
		    ++do_write;
		if(!*p)
		{
		    if( ++n >= argc )
			usage();
		    p = argv[n];
		}
		sscanf( p, "%x:%x", &ee_addr, &ee_data );
		if( ee_addr > 0x3F || ee_data > 0xFFFF )
		{
		    fprintf( stderr, "Address must be in the range 00-3F, data 0000-FFFF\n" );
		    exit( -1 );
		}
		break;
	   default:
		usage();
		break;
	}
    }

    if (verbose)
	printf(version_msg);

	/* Get access to all of I/O space. */
    if (iopl(3) < 0)
    {
	perror("Network adapter diagnostic: iopl()");
	fprintf(stderr, "This program must be run as root.\n");
	return 2;
    }

    /* Try to read a likely port_base value from /proc/pci. */
    if ( scan_proc_pci(card_num) == 0)
    {
	fprintf(stderr, "Unable to find a recognized card in /proc/pci.\n");
	return ENODEV;
    }
    return 0;
}

/* Generic (all PCI diags) code to find cards. */

static char bogus_iobase[] =
"This chip has not been assigned a valid I/O address, and will not function.\n"
" If you have warm-booted from another operating system, a complete \n"
" shut-down and power cycle may restore the card to normal operation.\n";

static char bogus_irq[] =
"This chip has not been assigned a valid IRQ, and will not function.\n"
" This must be fixed in the PCI BIOS setup.  The device driver has no way\n"
" of changing the PCI IRQ settings.\n"
" See  http://www.scyld.com/expert/irq-conflict.html  for more information.\n";

static int scan_proc_bus_pci(int card_num)
{
    int card_cnt = 0, chip_idx = 0;
    int port_base;
    char buffer[514];
    unsigned int pci_bus, pci_devid, irq, pciaddr0, pciaddr1;
    int i;
    FILE *fp = fopen("/proc/bus/pci/devices", "r");

    if (fp == NULL) {
	if (debug) fprintf(stderr, "Failed to open /proc/bus/pci/devices.\n");
	return -1;
    }
	
    while (fgets(buffer, sizeof(buffer), fp))
    {
	if (debug > 1)
	    fprintf(stderr, " Parsing line -- %s", buffer);
	if (sscanf(buffer, "%x %x %x %x %x", &pci_bus, &pci_devid, &irq, &pciaddr0, &pciaddr1) <= 0)
	     break;
	for (i = 0; pcidev_tbl[i].vendor; i++)
	{
	    if ((pci_devid >> 16) != pcidev_tbl[i].vendor
		|| (pci_devid & pcidev_tbl[i].device_mask) != pcidev_tbl[i].device)
		continue;
	    chip_idx = i;
	    card_cnt++;
	    /* Select the I/O address. */
	    port_base = pciaddr0 & 1  ?  pciaddr0 & ~1 : pciaddr1 & ~1;
	    if (card_num == 0 || card_num == card_cnt)
	    {
		printf("Index #%d: Found a %s adapter at %#x.\n",
		   card_cnt, pcidev_tbl[chip_idx].part_name, port_base);
		if (irq == 0  || irq == 255)
		    printf(bogus_irq);
		if (port_base)
		    pcidev_tbl[chip_idx].diag_func(0,0,port_base, i);
	    	else
		    printf(bogus_iobase);
		break;
	    }
	}
    }
    fclose(fp);
    return card_cnt;
}

static int scan_proc_pci(int card_num)
{
    int card_cnt = 0, chip_idx = 0;
    char chip_name[40];
    FILE *fp;
    int port_base;

    if ((card_cnt = scan_proc_bus_pci(card_num)) >= 0)
	return card_cnt;
    card_cnt = 0;

    fp = fopen("/proc/pci", "r");
    if (fp == NULL)
	return 0;
    {
	char buffer[514];
	int pci_bus, pci_device, pci_function, vendor_id, device_id;
	int state = 0;
	if (debug) printf("Done open of /proc/pci.\n");
	while (fgets(buffer, sizeof(buffer), fp))
	{
	    if (debug > 1)
		fprintf(stderr, " Parse state %d line -- %s", state, buffer);
	    if (sscanf(buffer, " Bus %d, device %d, function %d",
					   &pci_bus, &pci_device, &pci_function) > 0)
	    {
		chip_idx = 0;
		state = 1;
		continue;
	    }
	    if (state == 1)
	    {
		if (sscanf(buffer, " Ethernet controller: %39[^\n]", chip_name) > 0)
		{
		    int i;
		    if (debug)
			printf("Named ethernet controller %s.\n", chip_name);
		    for (i = 0; pcidev_tbl[i].part_name; i++)
			if (pcidev_tbl[i].proc_pci_name
				  && strncmp(pcidev_tbl[i].proc_pci_name,
					 chip_name, strlen(pcidev_tbl[i].proc_pci_name)) == 0)
			{
			    state = 2;
			    chip_idx = i;
			    continue;
			}
		    continue;
		}
		/* Handle a /proc/pci that does not recognize the card. */
		if (sscanf(buffer, " Vendor id=%x. Device id=%x", &vendor_id, &device_id) > 0)
		{
		    int i;
		    if (debug)
			printf("Found vendor 0x%4.4x device ID 0x%4.4x.\n", vendor_id, device_id);
		    for (i = 0; pcidev_tbl[i].vendor; i++)
			if (vendor_id == pcidev_tbl[i].vendor  &&
				(device_id & pcidev_tbl[i].device_mask) == pcidev_tbl[i].device)
			    break;
		    if (pcidev_tbl[i].vendor == 0)
			continue;
		    chip_idx = i;
		    state = 2;
		}
	    }
	    if (state == 2)
	    {
		if (sscanf(buffer, "  I/O at %x", &port_base) > 0)
		{
		    card_cnt++;
		    state = 3;
		    if (card_num == 0 || card_num == card_cnt)
		    {
			printf("Index #%d: Found a %s adapter at %#x.\n",
				   card_cnt, pcidev_tbl[chip_idx].part_name, port_base);
			if (port_base)
			    pcidev_tbl[chip_idx].diag_func(vendor_id, device_id, port_base, chip_idx);
			else
			    printf(bogus_iobase);
		    }
		}
	    }
	}
    }
    fclose(fp);
    return card_cnt;
}

//* Chip-specific section. */

/* The National Semiconductor DP83810 series. */


/* Offsets to the DP83815 registers.*/
enum register_offsets {
	ChipCmd=0x00,    ChipConfig=0x04, EECtrl=0x08, TestControl=0x0C,
	IntrStatus=0x10, IntrMask=0x14,   IntrEnable=0x18,
	TxRingPtr=0x20,  TxConfig=0x24,
	RxRingPtr=0x30,  RxConfig=0x34,
	WOLCmd=0x40,     PauseCmd=0x44,   RxFilterAddr=0x48, RxFilterData=0x4C,
	BootRomAddr=0x50,BootRomData=0x54,StatsCtrl=0x5C, StatsData=0x60,
	RxPktErrs=0x60,  RxMissed=0x68,   RxCRCErrs=0x64,
};

/* Read the EEPROM 
   The EEPROM code is for the common 93c06/46 EEPROMs with 6 bit addresses. */

/* Delay between EEPROM clock transitions.
   This flushes the write buffer to prevent quick double-writes.
*/
#define eeprom_delay(ee_addr)	inl(ee_addr); inl(ee_addr)

enum EEPROM_Ctrl_Bits {
	EE_ShiftClk=0x04, EE_DataIn=0x01, EE_ChipSelect=0x08, EE_DataOut=0x02,
};
#define EE_Write0 (EE_ChipSelect)
#define EE_Write1 (EE_ChipSelect | EE_DataIn)

/* The EEPROM commands always start with 01.. preamble bits.
   Commands are prepended to the variable-length address. */
enum EEPROM_Cmds { EE_WriteCmd=5, EE_ReadCmd=6, EE_EraseCmd=7, };

/* Values read from the EEPROM, and a new image to write. */
#define EEPROM_SIZE 256
static unsigned short eeprom_contents[EEPROM_SIZE];
static unsigned short new_ee_contents[EEPROM_SIZE];
#define EEPROM_SA_OFFSET	7
#define EE_OFFSET		EECtrl	/* Register offset in I/O space. */

/* Execute a generic EEPROM command.
   Return all data output from the EEPROM, and thus may be used for
   EEPROM sizing, read, erase or write. */
static int do_eeprom_cmd(long ioaddr, int cmd, int cmd_len)
{
    unsigned retval = 0;
    long    ee_addr = ioaddr + EE_OFFSET;

    if (debug > 1)
	printf(" EEPROM op 0x%x: ", cmd);

    outl(EE_ChipSelect | EE_ShiftClk, ee_addr);

    /* Shift the command bits out. */
    do {
	short dataval = (cmd & (1 << cmd_len)) ? EE_Write1 : EE_Write0;
	outl(dataval, ee_addr);
	eeprom_delay(ee_addr);
	if (debug > 2)
	    printf("%X", inl(ee_addr) & 15);
	outl(dataval | EE_ShiftClk, ee_addr);
	eeprom_delay(ee_addr);
	retval = (retval << 1) | ((inl(ee_addr) & EE_DataOut) ? 1 : 0);
    } while (--cmd_len >= 0);
    outl(EE_ChipSelect, ee_addr);

    /* Terminate the EEPROM access. */
    outl(0, ee_addr);
    if (debug > 1)
	printf(" EEPROM result is 0x%5.5x.\n", retval);
    return retval;
}

/* Wait for the EEPROM to finish what it is doing. */
static int eeprom_busy_poll(long ee_ioaddr)
{
	int i;
	outl(EE_ChipSelect, ee_ioaddr);
	for (i = 0; i < 10000; i++)			/* Typical 2000 ticks */
		if (inl(ee_ioaddr) & EE_DataOut)
			break;
	return i;
}

static int bit_reverse_16(int val)
{
	int i, result = 0;
	for (i = 0; i < 16; i++) {
		result = (result << 1) | (val & 1);
		val >>= 1;
	}
	return result;
}

/* The abstracted functions for EEPROM access. */
static int read_eeprom(long ioaddr, int location, int addr_len)
{
	return do_eeprom_cmd(ioaddr, ((EE_ReadCmd << addr_len) | location) << 16,
						 3 + addr_len + 16) & 0xffff;
}

static void write_eeprom(long ioaddr, int index, int value, int addr_len)
{
	long ee_ioaddr = ioaddr + EE_OFFSET;
	int i;

	/* Poll for previous op finished. */
	eeprom_busy_poll(ee_ioaddr);

	/* Enable programming modes. */
	do_eeprom_cmd(ioaddr, (0x4f << (addr_len-4)), 3 + addr_len);
	/* Do the actual write. */ 
	do_eeprom_cmd(ioaddr,
				  (((EE_WriteCmd<<addr_len) | index) << 16)
				  | bit_reverse_16(value),
				  3 + addr_len + 16);
	i = eeprom_busy_poll(ee_ioaddr);
	if (debug)
		printf(" Write finished after %d ticks.\n", i);
	/* Disable programming.  Note: this command is not instantaneous, so
	   we check for busy before the next op. */
	do_eeprom_cmd(ioaddr, (0x40 << (addr_len-4)), 3 + addr_len);
	eeprom_busy_poll(ee_ioaddr);
}

int natsemi_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
{
    unsigned int chip_active = inl(ioaddr + IntrEnable);
    int ee_addr_len = 6, eeprom_size = 64;
    int i, csum;

    if (chip_active == 0xffffffff) {
	printf(" * A recognized chip has been found, but it does not "
	   "appear to exist in\n * I/O space.\n");
	    return 1;
    }

    ee_addr_len = do_eeprom_cmd(ioaddr, EE_ReadCmd << (6+16), 3 + 6 + 16) & 0x10000 ? 8 : 6;
    eeprom_size = 1 << ee_addr_len;
    printf(" EEPROM address length %d, %d words.\n", ee_addr_len, eeprom_size);

    // Read the EEPROM contents and maybe Bit-reverse the result.
    csum = 0;
    for (i = 0; i < eeprom_size; i++)
    {
	int d;

	d = read_eeprom(ioaddr, i, ee_addr_len);
	csum += d;
	if( reverse )
	    eeprom_contents[i] = bit_reverse_16(d);
	else
	    eeprom_contents[i] = d;
	new_ee_contents[i] = d;
    }
    csum &= 0xFFFF;

    printf("EEPROM contents (%d words):", eeprom_size);
    for (i = 0; i < eeprom_size; i += 8)
    {
	int j;
	printf("\n0x%2.2x: ", i);
	for (j = 0; j < 8; j++)
	    printf(" %4.4x", eeprom_contents[i + j]);
    }
    printf("\nOverall Checksum 0x%04x\n", csum );

    // See if we're to change anything
    if( ee_addr < 0x40 )
    {
	printf( "Request to change address %02X to %04X\n", ee_addr, ee_data );
	new_ee_contents[ee_addr] = ee_data;
	csum = 0;
	for( i=0; i<eeprom_size-1; ++i )
	    csum += new_ee_contents[i];
	new_ee_contents[i] = (-csum) & 0xFFFF;

	for (i = 0; i < eeprom_size; i++)
	{
	    if (new_ee_contents[i] != eeprom_contents[i])
	    {
		if (do_write)
		{
		    printf( "Changing address %02X to %04X\n", i, new_ee_contents[i] );
		    write_eeprom(ioaddr, i, bit_reverse_16(new_ee_contents[i]), ee_addr_len);
		}
		else
		    printf("Would write %4.4x to replace %4.4x at %02X.\n",
				   new_ee_contents[i], eeprom_contents[i], i);
	    }
	}
    }
#ifdef ZZZ
//write_eeprom(ioaddr, 0x18, 0x2800, ee_addr_len);
//write_eeprom(ioaddr, 0x3F, 0xC6F2, ee_addr_len);

	if (show_eeprom > 1) {
		int csum = 0;
		printf("EEPROM contents (%d words):", eeprom_size);
		for (i = 0; i < eeprom_size; i += 8) {
			int j;
			printf("\n0x%2.2x: ", i);
			for (j = 0; j < 8; j++){
				csum += eeprom_contents[i+j];
				printf(" %4.4x", eeprom_contents[i + j]);
			}
			if (show_eeprom > 2) {
				printf("  ");
				for (j = 0; j < 8; j++) {
					int ew = eeprom_contents[i + j];
					printf("%c%c",
						   isprint(ew & 0xff) ? ew & 0xff : '_',
						   isprint(ew >>   8) ? ew >> 8   : '_' );
				}
			}
		}
		printf("\nOverall Checksum 0x%04x\n", csum );

	}
#endif
    return 0;
}

/*
 * Local variables:
 *  compile-command: "cc -O -Wall -o evo_eeprom evo_eeprom.c"
 *  tab-width: 4
 *  c-indent-level: 4
 *  c-basic-offset: 4
 * End:
 */
